# ***** BEGIN GPL LICENSE BLOCK ***** 
# 
# This program is free software; you can redistribute it and/or 
# modify it under the terms of the GNU General Public License 
# as published by the Free Software Foundation; either version 2 
# of the License, or (at your option) any later version. 
# 
# This program is distributed in the hope that it will be useful, 
# but WITHOUT ANY WARRANTY; without even the implied warranty of 
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
# GNU General Public License for more details. 
# 
# You should have received a copy of the GNU General Public License 
# along with this program; if not, write to the Free Software Foundation, 
# Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA. 
# 
# ***** END GPL LICENCE BLOCK *****

import os
import bpy

""" 
Name: 'ASCII Scene Importer' 
Blender: 2.53 Beta
Group: 'Import' 
Tooltip: 'ASCII Scene import (*.ase)' 
""" 
__author__ = "Richard Bartlett" 
__version__ = "0.03" 

#TODO:

class ase:
    def __init__(self):
        self.version = ''
        self.comment = ''
        self.scene = scene()
        self.materials = []             # list of material instances
        self.geomobjects = []           # list of geomobject instances

class scene:
    def __init__(self):
        self.filename = ''
        self.firstframe = 0
        self.lastframe = 0
        self.framespeed = 0
        self.ticsperframe = 0
        self.backgroundstatic = 0.0, 0.0, 0.0
        self.ambientstatic = 0.0, 0.0, 0.0

class material:
    def __init__(self):
        self.name = ''
        self.matclass = ''
        self.ambient = 0.0, 0.0, 0.0
        self.diffuse = 0.0, 0.0, 0.0
        self.specular = 0.0, 0.0, 0.0
        self.shine = 0.0
        self.shinestrength = 0.0
        self.transparency = 0.0
        self.wiresize = 0.0
        self.numsubmtls = 0
        self.shading = ''
        self.xpfalloff = 0.0
        self.xptype = ''
        self.falloff = ''
        self.soften = False
        self.diffusemap = None
        self.submtls = []
        self.selfillum = 0.0
        
class diffusemap:
    def __init__(self):
        self.name = ''
        self.mapclass = ''
        self.subno = 0
        self.amount = 0.0
        self.bitmap = ''
        self.type = ''
        self.uoffset = 0.0
        self.voffset = 0.0
        self.utiling = 0.0
        self.vtiling = 0.0
        self.angle = 0.0
        self.blur = 0.0
        self.bluroffset = 0.0
        self.noiseamt = 0.0
        self.noisesize = 0.0
        self.noiselevel = 0
        self.noisephase = 0.0
        self.bitmapfilter = ''
        

class geomobject:
    def __init__(self):
        self.name = ''
        self.tm = tm()
        self.mesh = mesh()
        self.motionblur = 0
        self.castshadow = 0
        self.recvshadow = 0
        self.materialref = 0
        
class tm:
    def __init__(self):
        self.name = ''
        self.inheritpos = 0, 0, 0
        self.inheritrot = 0, 0, 0
        self.inheritscale = 0, 0, 0
        self.tmrow0 = 0.0, 0.0, 0.0
        self.tmrow1 = 0.0, 0.0, 0.0
        self.tmrow2 = 0.0, 0.0, 0.0
        self.tmrow3 = 0.0, 0.0, 0.0
        self.tmpos = 0.0, 0.0, 0.0
        self.tmrotaxis = 0.0, 0.0, 0.0
        self.tmrotangle = 0.0
        self.tmscale = 0.0, 0.0, 0.0
        self.tmscaleaxis = 0.0, 0.0, 0.0
        self.tmscaleaxisang = 0.0

class mesh:
    def __init__(self):
        self.timevalue = 0
        self.numvertex = 0
        self.numfaces = 0
        self.vertexlist = []
        self.facelist = []
        self.numtvertex = 0
        self.tvertlist = []
        self.numtvfaces = 0
        self.tfacelist = []
        self.numcvertex = 0
        self.cvertlist = []
        self.numcvfaces = 0
        self.cfacelist = []
        self.meshnormals = []
        
class vertex:
    def __init__(self):
        self.coor = 0.0, 0.0, 0.0
        
class face:
    def __init__(self):
        self.verts = 0, 0, 0
        self.edgeflags = 0, 0, 0
        self.smoothgroups = []
        self.mtlid = 0

class tvert:
    def __init__(self):
        self.uvw = 0.0, 0.0, 0.0

class tface:
    def __init__(self):
        self.tverts = 0, 0, 0

class cvert:
    def __init__(self):
        self.rgb = 0.0, 0.0, 0.0

class cface:
    def __init__(self):
        self.cverts = 0, 0, 0
        
class vertnormal:
    def __init__(self):
        self.index = 0
        self.normal = 0.0, 0.0, 0.0

class facenormal:
    def __init__(self):
        self.normal = 0.0, 0.0, 0.0
        self.verts = []
        
class datastack:
    def __init__(self, file):
        self.data = [x.strip() for x in list(file)]         #remove leading and trailing whitespace
        self.data = self.data[::-1]                         #reverse order
        self.data = [x for x in self.data if len(x) > 0]    #remove blank lines
        self.numlines = len(self.data)
        
class tokenError(Exception):
    def __init__(self, token, function):
        self.token = token
        self.function = function
    def __str__(self):
        return repr(self.token, self.function)

def importASE(filename):
    print('*** ASE Scene Importer v' + __version__ + ' ***')
    print('Loading', filename)
    model = ase()
    try:
        file = open(filename, 'rU')
    except IOError:
        print('ERROR: The file does not exist. Aborting.')
    else:
        asestack = datastack(file)
        file.close()
        if asestack.numlines == 0:
            print('ERROR: The file contains no data. Aborting.')
            return
        try:
            readHeader(asestack, model)
            readScene(asestack, model)
            readMaterialList(asestack, model)
            readGeomObjects(asestack, model)
            print('Success.')
            return model
        except tokenError as error:
             print('ERROR:', error.function, 'encountered unexpected token', error.token, 'on line', asestack.numlines - len(asestack.data), 'of file', filename)
  
def readHeader(stack, model):
    while len(stack.data) > 0:
        line = stack.data.pop()
        if line.startswith('*3DSMAX_ASCIIEXPORT'):
            model.version = line.split()[1]
        elif line.startswith('*COMMENT'):
            model.comment = line[10:-1]
            break
        else:
            raise tokenError(line.split()[0], 'readHeader')
            
def readScene(stack, model):
    line = stack.data.pop()
    if line.startswith('*SCENE {'):
        while len(stack.data) > 0:
            line = stack.data.pop()
            if line.startswith('}'):
                return
            elif line.startswith('*SCENE_FILENAME'):
                model.scene.filename = line.split()[1][1:-1]
            elif line.startswith('*SCENE_FIRSTFRAME'):
                model.scene.firstframe = int(line.split()[1])
            elif line.startswith('*SCENE_LASTFRAME'):
                model.scene.lastframe = int(line.split()[1])
            elif line.startswith('*SCENE_FRAMESPEED'):
                model.scene.framespeed = int(line.split()[1])
            elif line.startswith('*SCENE_TICKSPERFRAME'):
                model.scene.ticksperframe = int(line.split()[1])
            elif line.startswith('*SCENE_BACKGROUND_STATIC'):
                model.scene.backgroundstatic = [float(x) for x in line.split()[1:]]
            elif line.startswith('*SCENE_AMBIENT_STATIC'):
                model.scene.ambientstatic = [float(x) for x in line.split()[1:]]
            else:
                raise tokenError(line.split()[0], 'readScene')
    else:
        raise tokenError(line.split()[0], 'readScene')

def readMaterialList(stack, model):
    line = stack.data.pop()
    if line.startswith('*MATERIAL_LIST {'):
        line = stack.data.pop()
        if line.startswith('*MATERIAL_COUNT'):
            count = int(line.split()[1])
            for x in range(count):
                stack.data.pop()
                model.materials.append(readMaterial(stack, model))
            line = stack.data.pop() # should be closing bracket            
        else:
            raise tokenError(line.split()[0], 'readMaterialList')
    else:
        raise tokenError(line.split()[0], 'readMaterialList')

def readMaterial(stack, model):
    mat = material()
    while len(stack.data) > 0:
        line = stack.data.pop()
        if line.startswith('}'):
            return mat
        elif line.startswith('*MATERIAL_NAME'):
            mat.name = line[line.index('"')+1:-1]
        elif line.startswith('*MATERIAL_CLASS'):
            mat.matclass = line[line.index('"')+1:-1]
        elif line.startswith('*MATERIAL_AMBIENT'):
            mat.ambient = [float(x) for x in line.split()[1:]]
        elif line.startswith('*MATERIAL_DIFFUSE'):
            mat.diffuse = [float(x) for x in line.split()[1:]]
        elif line.startswith('*MATERIAL_SPECULAR'):
            mat.specular = [float(x) for x in line.split()[1:]]
        elif line.startswith('*MATERIAL_SHINE'):
            mat.shine = float(line.split()[1])
        elif line.startswith('*MATERIAL_SHINESTRENGTH'):
            mat.shinestrength = float(line.split()[1])
        elif line.startswith('*MATERIAL_TRANSPARENCY'):
            mat.transparency = float(line.split()[1])
        elif line.startswith('*MATERIAL_WIRESIZE'):
            mat.wiresize = float(line.split()[1])
        elif line.startswith('*MATERIAL_SHADING'):
            mat.shading = line.split()[1]
        elif line.startswith('*MATERIAL_XP_FALLOFF'):
            mat.xpfalloff = float(line.split()[1])
        elif line.startswith('*MATERIAL_SELFILLUM'):
            mat.selfillum = float(line.split()[1])
        elif line.startswith('*MATERIAL_FALLOFF'):
            mat.falloff = line.split()[1]
        elif line.startswith('*MATERIAL_XP_TYPE'):
            mat.xptype = line.split()[1]
        elif line.startswith('*MAP_DIFFUSE {'):
            mat.diffusemap = readDiffusemap(stack, model)
        elif line.startswith('*MATERIAL_SOFTEN'):
            mat.soften = True
        elif line.startswith('*NUMSUBMTLS'):
            mat.numsubmtls = int(line.split()[1])
        elif line.startswith('*SUBMATERIAL '):
            submat = readMaterial(stack, model)
            mat.submtls.append(submat)
        else:
            raise tokenError(line.split()[0], 'readMaterial')


def readDiffusemap(stack, model):
    diff = diffusemap()
    while len(stack.data) > 0:
        line = stack.data.pop()
        if line.startswith('}'):
            return diff
        elif line.startswith('*MAP_NAME'):
            diff.name = line[line.index('"')+1:-1]
        elif line.startswith('*MAP_CLASS'):
            diff.mapclass = line[line.index('"')+1:-1]
        elif line.startswith('*MAP_SUBNO'):
            diff.subno = int(line.split()[1])
        elif line.startswith('*MAP_AMOUNT'):
            diff.amount = float(line.split()[1])
        elif line.startswith('*BITMAP '):
            diff.bitmap = line[line.index('"')+1:-1]
            if ('\\base\\' in diff.bitmap) or ('/base/' in diff.bitmap):
                diff.bitmap = diff.bitmap.replace('/', '\\')
                diff.bitmap = diff.bitmap.split('\\base\\')[1]
        elif line.startswith('*MAP_TYPE'):
            diff.type = line.split()[1]
        elif line.startswith('*UVW_U_OFFSET'):
            diff.uoffset = float(line.split()[1])
        elif line.startswith('*UVW_V_OFFSET'):
            diff.voffset = float(line.split()[1])
        elif line.startswith('*UVW_U_TILING'):
            diff.utiling = float(line.split()[1])
        elif line.startswith('*UVW_V_TILING'):
            diff.vtiling = float(line.split()[1])
        elif line.startswith('*UVW_ANGLE'):
            diff.uangle = float(line.split()[1])
        elif line.startswith('*UVW_BLUR'):
            diff.blur = float(line.split()[1])
        elif line.startswith('*UVW_BLUR_OFFSET'):
            diff.bluroffset = float(line.split()[1])
        elif line.startswith('*UVW_NOUSE_AMT'):  # IS THIS REALLY MISSPELLED IN THE SPEC?
            diff.noiseamt = float(line.split()[1])
        elif line.startswith('*UVW_NOISE_SIZE'):
            diff.noisesize = float(line.split()[1])
        elif line.startswith('*UVW_NOISE_LEVEL'):
            diff.noiselevel = int(line.split()[1])
        elif line.startswith('*UVW_NOISE_PHASE'):
            diff.noisephase = float(line.split()[1])
        elif line.startswith('*BITMAP_FILTER'):
            diff.bitmapfilter = line.split()[1]
        else:
            raise tokenError(line.split()[0], 'readDiffusemap')

def readGeomObjects(stack, model):
    while len(stack.data) > 0:
        line = stack.data.pop()
        if line.startswith('*GEOMOBJECT {'):
            model.geomobjects.append(readGeomObject(stack, model))
        else:
            raise tokenError(line.split()[0], 'readGeomObjects')
            
def readGeomObject(stack, model):
    geo = geomobject()
    while len(stack.data) > 0:
        line = stack.data.pop()
        if line.startswith('}'):
            return geo
        elif line.startswith('*NODE_NAME'):
            geo.name = line[line.index('"')+1:-1]
        elif line.startswith('*NODE_TM'):
            geo.tm = readTm(stack, model)
        elif line.startswith('*MESH '):
            geo.mesh = readMesh(stack, model)
        elif line.startswith('*PROP_MOTIONBLUR'):
            geo.motionblur = int(line.split()[1])
        elif line.startswith('*PROP_CASTSHADOW'):
            geo.castshadow = int(line.split()[1])
        elif line.startswith('*PROP_RECVSHADOW'):
            geo.recvshadow = int(line.split()[1])
        elif line.startswith('*MATERIAL_REF'):
            geo.materialref = int(line.split()[1])
        else:
            raise tokenError(line.split()[0], 'readGeomObject')
            
def readTm(stack, model):
    tmdata = tm()
    while len(stack.data) > 0:
        line = stack.data.pop()
        if line.startswith('}'):
            return tmdata
        elif line.startswith('*NODE_NAME'):
            tmdata.name = line[line.index('"')+1:-1]
        elif line.startswith('*INHERIT_POS'):
            tmdata.inheritpos = [int(x) for x in line.split()[1:]]
        elif line.startswith('*INHERIT_ROT'):
            tmdata.inheritrot = [int(x) for x in line.split()[1:]]
        elif line.startswith('*INHERIT_SCL'):
            tmdata.inheritscl = [int(x) for x in line.split()[1:]]
        elif line.startswith('*TM_ROW0'):
            tmdata.tmrow0 = [float(x) for x in line.split()[1:]]
        elif line.startswith('*TM_ROW1'):
            tmdata.tmrow1 = [float(x) for x in line.split()[1:]]
        elif line.startswith('*TM_ROW2'):
            tmdata.tmrow2 = [float(x) for x in line.split()[1:]]
        elif line.startswith('*TM_ROW3'):
            tmdata.tmrow3 = [float(x) for x in line.split()[1:]]
        elif line.startswith('*TM_POS'):
            tmdata.tmpos = [float(x) for x in line.split()[1:]]
        elif line.startswith('*TM_ROTAXIS'):
            tmdata.tmrotaxis = [float(x) for x in line.split()[1:]]
        elif line.startswith('*TM_ROTANGLE'):
            tmdata.tmrotangle = float(line.split()[1])
        elif line.startswith('*TM_SCALE'):
            tmdata.tmscale = [float(x) for x in line.split()[1:]]
        elif line.startswith('*TM_SCALEAXIS'):
            tmdata.tmscaleaxis = [float(x) for x in line.split()[1:]]
        elif line.startswith('*TM_SCALEAXISANG'):
            tmdata.tmscaleaxisang = float(line.split()[1])
        else:
            raise tokenError(line.split()[0], 'readTm')
            
def readMesh(stack, model):
    meshdata = mesh()
    while len(stack.data) > 0:
        line = stack.data.pop()
        if line.startswith('}'):
            return meshdata
        elif line.startswith('*TIMEVALUE'):
            meshdata.timevalue = int(line.split()[1])
        elif line.startswith('*MESH_NUMVERTEX'):
            meshdata.numvertex = int(line.split()[1])
        elif line.startswith('*MESH_NUMFACES'):
            meshdata.numfaces = int(line.split()[1])
        elif line.startswith('*MESH_VERTEX_LIST'):
            for x in range(meshdata.numvertex):
                vertexdata = vertex()
                line = stack.data.pop()
                vertexdata.coor = [float(x) for x in line.split()[2:]]
                meshdata.vertexlist.extend(vertexdata.coor)
            stack.data.pop()
        elif line.startswith('*MESH_FACE_LIST'):
            for x in range(meshdata.numfaces):
                facedata = face()
                line = stack.data.pop().split()
                facedata.verts = int(line[3]), int(line[5]), int(line[7])
                facedata.edgeflags = int(line[9]), int(line[11]), int(line[13])
                facedata.smoothgroups = line[::-1][2]
                if facedata.smoothgroups == '*MESH_SMOOTHING':
                    facedata.smoothgroups = ''
                face.mtlid = int(line[::-1][0])
                meshdata.facelist.append(facedata)
            stack.data.pop()
        elif line.startswith('*MESH_NUMTVERTEX'):
            meshdata.numtvertex = int(line.split()[1])
        elif line.startswith('*MESH_TVERTLIST'):
            for x in range(meshdata.numtvertex):
                tvertdata = tvert()
                line = stack.data.pop()
                tvertdata.uvw = [float(x) for x in line.split()[2:]]
                meshdata.tvertlist.append(tvertdata)
            stack.data.pop()
        elif line.startswith('*MESH_NUMTVFACES'):
            meshdata.numtvfaces = int(line.split()[1])
        elif line.startswith('*MESH_TFACELIST'):
            for x in range(meshdata.numtvfaces):
                tfacedata = tface()
                line = stack.data.pop()
                tfacedata.tverts = [int(x) for x in line.split()[2:]]
                meshdata.tfacelist.append(tfacedata)
            stack.data.pop()
        elif line.startswith('*MESH_NORMALS'):
            for x in range(meshdata.numfaces):
                facenormaldata = facenormal()
                facenormaldata.normal = [float(x) for x in stack.data.pop().split()[2:]]
                for y in range(3):
                    temp = stack.data.pop().split()[1:]
                    vertnormaldata = vertnormal()
                    vertnormaldata.index = int(temp[0])
                    vertnormaldata.normal = [float(x) for x in temp[1:]]
                    facenormaldata.verts.append(vertnormaldata)
                meshdata.meshnormals.append(facenormaldata)
            stack.data.pop()
        elif line.startswith('*MESH_NUMCVERTEX'):
            meshdata.numcvertex = int(line.split()[1])
        elif line.startswith('*MESH_CVERTLIST'):
            for x in range(meshdata.numcvertex):
                cvertdata = cvert()
                line = stack.data.pop()
                cvertdata.rgb = [float(x) for x in line.split()[2:]]
                meshdata.cvertlist.append(cvertdata)
            stack.data.pop()
        elif line.startswith('*MESH_NUMCVFACES'):
            meshdata.numcvfaces = int(line.split()[1])
        elif line.startswith('*MESH_CFACELIST'):
            for x in range(meshdata.numcvfaces):
                cfacedata = cface()
                line = stack.data.pop()
                cfacedata.cverts = [int(x) for x in line.split()[2:]]
                meshdata.cfacelist.append(cfacedata)
            stack.data.pop()
        else:
            raise tokenError(line.split()[0], 'readMesh')

def applyData(model, scale):
    import os
    
    for x in model.materials:
        image = bpy.ops.image.new(name = 'Image - ' + str(model.materials.index(x)))
        image = bpy.data.images.get('Image - ' + str(model.materials.index(x)))
        image.source = 'FILE'
        image.filepath = x.diffusemap.bitmap
        texture = bpy.data.textures.new('Texture - ' + str(model.materials.index(x)))
        texture.type = 'IMAGE'
        texture = texture.recast_type()
        texture.image = image
        material = bpy.data.materials.new('Material - ' + str(model.materials.index(x)))
        material.add_texture(texture)
        material.diffuse_color = x.diffuse
        material.specular_color = x.specular
        material.specular_hardness = int(x.shine * 511)
        if x.shinestrength > 0:
            material.specular_intensity = x.shinestrength
        if x.transparency > 0:
            material.transparency = True
            material.alpha = x.transparency
        if x.shading in ['Blinn', 'Phong']:
            material.specular_shader = x.shading.upper()
        if x.selfillum > 0:
            material.emit = x.selfillum

    for x in model.geomobjects:
        mesh = bpy.data.meshes.new(x.name)
        mesh.add_geometry(x.mesh.numvertex, 0, x.mesh.numfaces)
        mesh.verts.foreach_set("co",x.mesh.vertexlist)
        mesh.faces.foreach_set("verts_raw", [y for z in x.mesh.facelist for y in [z.verts[0], z.verts[1], z.verts[2], z.verts[0]]])
        
        if x.mesh.numtvfaces > 0:
            mesh.add_uv_texture()
            for i in range(x.mesh.numtvfaces):
                mesh.uv_textures[0].data[i].uv1 = x.mesh.tvertlist[x.mesh.tfacelist[i].tverts[0]].uvw[:-1]
                mesh.uv_textures[0].data[i].uv2 = x.mesh.tvertlist[x.mesh.tfacelist[i].tverts[1]].uvw[:-1]
                mesh.uv_textures[0].data[i].uv3 = x.mesh.tvertlist[x.mesh.tfacelist[i].tverts[2]].uvw[:-1]
                
        if x.mesh.numcvfaces > 0:
            mesh.add_vertex_color()
            for i in range(x.mesh.numcvfaces):
                mesh.vertex_colors[0].data[i].color1 = x.mesh.cvertlist[x.mesh.cfacelist[i].cverts[0]].rgb
                mesh.vertex_colors[0].data[i].color2 = x.mesh.cvertlist[x.mesh.cfacelist[i].cverts[1]].rgb
                mesh.vertex_colors[0].data[i].color3 = x.mesh.cvertlist[x.mesh.cfacelist[i].cverts[2]].rgb
        
        mesh.update()
        
        # grab scene context
        scene = bpy.context.scene
        
        # deselect all objects in scene
        for ob in scene.objects:
            ob.select = False
        
        # create new object, populate it with mesh data, and link it to scene
        ob_new = bpy.data.objects.new(x.name, mesh)
        scene.objects.link(ob_new)
        ob_new.select = True
        ob_new.data.add_material(bpy.data.materials.get('Material - '+ str(x.materialref)))
        bpy.context.scene.objects.active = ob_new
        bpy.ops.object.shade_smooth()
        bpy.ops.transform.resize(value=(scale, scale, scale))    

def test():
    samplepath = 'C:/Users/Richard/Desktop/blenderplugin/samples'
    print('*********************************************************')
    for filename in os.listdir(samplepath):
        if filename.endswith('.ASE') or filename.endswith('.ase'):
            model = importASE(os.path.join(samplepath, filename))
            applyData(model)

from bpy.props import *

class IMPORT_OT_ase(bpy.types.Operator):
    '''Load an Ascii Scene Export File'''
    bl_idname = "import_scene.ase"
    bl_label = "Import ASE"
  
    filepath = StringProperty(name="File Path", description="File path used for importing the ASE file", maxlen= 1024, default= "")
    ASE_SCALE = FloatProperty(name="Scale", description="Object scaling factor (default: 1/16)", min=0.01, max=1000.0, soft_min=0.01, soft_max=1000.0, default=(1/16))
    
    def execute(self, context):
        model = importASE(self.properties.filepath)
        applyData(model, self.properties.ASE_SCALE)
        return {'FINISHED'}
    
    def invoke(self, context, event):
        wm = context.manager
        wm.add_fileselect(self)
        return {'RUNNING_MODAL'}

menu_func = lambda self, context: self.layout.operator(IMPORT_OT_ase.bl_idname, text='Ascii Scene Export (.ase)')

def register():
    bpy.types.register(IMPORT_OT_ase)
    bpy.types.INFO_MT_file_import.append(menu_func)
    
def unregister():
    bpy.types.unregister(IMPORT_OT_ase)
    bpy.types.INFO_MT_file_import.remove(menu_func)

if __name__ == '__main__':
    register()